		(* manipulations concernant le groupe symtrique *)

(* la liste 1..n dont on a tout le temps besoin *)
let range n=
	let rec prefixe p q= 
		if p=q
		then [q]
		else p::prefixe (p+1) q in
	prefixe 1 n
;;

(* entre d'une permutation comme liste, transformation en une fonction de 
1..n->1..n*)
let cree_perm l=
	let tab = vect_of_list l in
	let f=fun k->tab.(k-1) in f
;;

(* affichage d'une permutation comme "matrice" Source et But *)
let affiche f n=
	let rec aff_liste l=match l with
		|[]->()
		|t::q->print_int t;print_string" "; aff_liste q in
	print_newline();
	aff_liste (range n);
	print_newline();
	aff_liste (map f (range n));
	print_newline()
;;

let perm_a=cree_perm [6;1;4;2;5;3];;
perm_a 4;;

#open "sort";;

let verifie f n =
	let valeurs=sort (prefix <) (map f (range n)) in 
	valeurs=(range n)
;;

verifie perm_a 5;;
verifie perm_a 4;;

let compose p q=fun k->p (q k);;

affiche (compose perm_a perm_a) 6;;
 
let range_vect n=vect_of_list (range n);;

	(* calcul de la permutation rciproque de p *)
let reciproque p n=
	if verifie p n
	then 	let bat = range_vect n in
		do_vect (fun k->bat.((p k)-1)<-k) (range_vect n);
		cree_perm (list_of_vect bat); 
	else failwith "non bijectif"
;;

affiche perm_a 6;;
let bat = range_vect 6 ;;
	do_vect (fun k->bat.((perm_a k)-1)<-k) (range_vect 6);;
list_of_vect bat;; 
perm_a 1;;

let perm_b= reciproque perm_a 6;;

affiche (compose perm_b perm_a) 6;;

		(* confection de permutations au hasard *)
#open "random";;
init (int_of_float (sys__time()));;;;

(* hasard va tre la liste des choix  faire pour crer la permutation, c'est la mme chose que la numration en base factorielle *)
let hasard n=
	let places=make_vect (n-1) 0 in 
	for k=0 to (n-2) do places.(k)<- int (k+1) done;
	places 
;;

(* Oh la vilaine rcursion non terminale! *)
let rec insere liste valeur rang=
	if rang=0 
	then valeur::liste
	else match liste with
		|[]->[valeur]
		|t::q->t::(insere q valeur (rang-1))
;;
	
(* ci-dessous un exemple *)
insere (list_of_vect (make_vect 5 0)) 8 3;;

(* aleaperm fournit une liste alatoire de 1..n, si on veut un tableau on fera suivre aleaperm de vect_of_list *)
let aleaperm n=
	let hasa=hasard n and debut=ref [1] in 
	for k= 2 to n 
	do debut:=insere !debut k hasa.(k-2) done;
	!debut
;;

(* ci-dessous un usage typique de where *)

let compose f g=
	if vect_length f <> vect_length g
	then failwith "Ah le bon gag!"
	else map_vect ff g
		where ff u=f.(u-1)
;;
	
(*dcomposition d'une permutation en produit de cycles  supports disjoints *)

(* ci-dessous de quoi extraire son premier cycle  une fonction *)
let premier_cycle q=
	let rec ajoute k resultat=
		let qk=q k in
		if mem qk resultat 
		then resultat else (ajoute qk (qk::resultat))
in ajoute 1 [1];;

(*map test (range 10);; premier_cycle test;;*)

let dec_cycles p n=
	let dejavu=make_vect (n+2) false 
in	let premier_cycle =function m->
		let rec ajoute k resultat=
			let pk=p k in
			if mem pk resultat 
			then resultat 
			else (dejavu.(pk)<-true;ajoute pk (pk::resultat))
		in dejavu.(m)<-true;ajoute m [m];
and 	suivant m=while dejavu.(!m) do incr m done;m
in 	let rec		ajoute_cycle k liste=
		if k=(n+1) 
		then liste 
		else ajoute_cycle !(suivant (ref k)) ((premier_cycle k) ::liste)
in ajoute_cycle 1 [];;

let pf=cree_perm [1;2;4;3;11;10;9;8;7;6;5];;

dec_cycles pf 11;;

(* quelques utilitaires fonctionnels avant la signature *)
let rec somme =function
	|[]->0
	|[u]->u
	|t::q->t+somme q
;;

let rec produit=function 
	|[]->1
	|[u]->u
	|t::q->t*produit q
;;

let rec somme_2 =function
	|[]->0
	|[u]->(u mod 2)
	|t::q->(t+somme_2 q) mod 2
;;

let rec produit_2=function 
	|[]->1
	|[u]->(u mod 2)
	|t::q->(t*produit q) mod 2
;;

let rec longueur=function
	|[]->0
	|t::q->1+longueur q
;;

let signature p n=
	let ll=dec_cycles p n and pair= function l->1+longueur l in
	produit_2 (map pair ll)
;;
	
(* affichage d'une table d'un groupe de permutations *)
(* la fonction associ_de x dfinie ci-dessous est tire de la page 142 de 
"Le Langage caml", elle ne sert  rien car elle est identique  la fonction 
assoc de caml standard *)
exception Pas_trouv;;
let rec associ_de x =function
	|[] -> raise Pas_trouv
	|(cl,valeur)::suite -> if x=cl then valeur else associ_de x suite;;
let mini_liste_test =[(5,"cinq");(8,"huit")];;
associ_de 5 mini_liste_test;;
associ_de 15 mini_liste_test;;

let S3=[([1;2;3],"I");([2;3;1],"c");([3;1;2],"c");([1;3;2],"t1");([3;2;1],"t2");([2;1;3],"t3")];;
assoc [1;3;2] S3;;

(* je recopie ce que le programme code de Gray a produit *)
let P3= [[|1; 2; 3|]; [|1; 3; 2|]; [|3; 1; 2|]; [|3; 2; 1|]; [|2; 3; 1|];  [|2; 1; 3|]];;
(* ... et je le remodle *)
let P3vect=vect_of_list(map list_of_vect P3);;

(* un affichage tabul des strings *)
let ptab s= let sl = string_length s in print_string(" "^s^(make_string (6-sl) ` `));; 

let tableS3= (* dans affiche_ligne p, p est un lment de P3vect *)
	let affiche_ligne p= let vp=vect_of_list p in 
		begin
			ptab (assoc p S3); 
			do_vect (fun z->ptab (assoc (list_of_vect (compose vp (vect_of_list z))) S3)) P3vect;
			print_newline()
		end
	in
	(* ligne d'en tte *)
	ptab "";
	do_vect ptab (map_vect (fun k->assoc k S3) P3vect);
	print_newline(); print_newline();
	(* les autres lignes *)
	do_vect affiche_ligne P3vect;print_newline()
 ;;




(*  faire
- composition FAIT
- rciproque FAIT
- cration de permutations au hasard FAIT
- dcomposition en cycles FAIT
- signature FAIT
- affichage de la table de Sn ou de celle de An FAIT
- compte du nombre d'inversions
- dire si deux permutations sont conjugues
- dtermination des parties stables
- code de gray pour Sn FAIT
- code de Gray pour An FAIT
- promenades dans le permutodre
- l'exo des jeux Mathmatiques et logiques 97 sur les retournements de sous listes FAIT
*)
