(** Le k-ime en O(n) selon AHU, p 291**)


(* Pour N<75, on a besoin de Mergesort *)
let rec fusion l1 l2 = match (l1,l2) with
	|[],_		   -> l2
	|_,[]		   -> l1
	|t1::q1,t2::q2 -> if t1<t2 then t1::(fusion q1 l2) else t2::(fusion l1 q2)
;;
let rec egrem = function
	|[]->([],[])
	|[a]->([a],[])
	|(a::b::z) -> let (u,v)= egrem z in (a::u,b::v)
;;
let rec mergesort= function
	| [] -> []
	| [a] -> [a]
	| liste -> let (x,y)=egrem liste in fusion (mergesort x) (mergesort y)
;;

(* Mergesort [2;5;8;9;1;0;6];; *)

let rec k_ime k = function
	|[] -> failwith "liste trop courte"
	|t::q -> if k=1 then t else k_ime (k-1) q
;;

(* Partition coupe en deux une liste  partir de pivot *)
let partition pivot =
	let rec parti liste_inf liste_sup nb_inf= function
		|[]	-> (liste_inf,liste_sup,nb_inf)
		|t::q -> if t<=pivot then parti (t::liste_inf) liste_sup (nb_inf+1) q
					   else parti liste_inf (t::liste_sup) nb_inf q
in	parti [] [] 0
;;

(* Minimdiane trouve la mdiane d'un qroupe de 5 lments en moins de 8 comparaisons (en fait 7 !!!) *)
let minimdiane liste = let med=list_length liste /2 +1
in	let rec fusion_mdiane l1 l2 k = match (l1,l2,k) with
		|[],_,_	     -> hd l2
		|_,[],_	     -> hd l1
		|t1::q1,t2::q2,1 -> min t1 t2
		|t1::q1,t2::q2,k -> if t1<t2 then fusion_mdiane q1 l2 (k-1) else fusion_mdiane l1 q2 (k-1)
		
in	match liste with
		|[]	   -> failwith "Erreur dans minimdiane : liste vide"
		|[a]	   -> a
		|[a;b]   -> a
		|a::b::q -> fusion_mdiane (mergesort [a;b]) (mergesort q) med
;;

(* minimdiane [1;2;3];;
minimdiane [1;2;3;4];;
minimdiane [1;2;3;4;5];; *)

(* Cette fonction va couper la liste en liste  5 lments *) 
let rec coupe_en_5 = function
	|[] -> []
	|a::b::c::d::e::q -> [a;b;c;d;e]::(coupe_en_5 q)
	|liste -> [liste]
;;

(* coupe_en_5 [1;2;3;4;5;6;7;8;];; *)

(* Selectionne va mettre en pratique l'algorithme dcrit dans AHU *) 
let rec selectionne k liste = let n=list_length liste in
	if n<75
	then k_ime k (mergesort liste)
	else let temp = map minimdiane (coupe_en_5 liste)
		  in let pivot= selectionne (n/10) temp
		  in let (liste_inf,liste_sup,nb_inf)=partition pivot liste
		  in if k<=nb_inf 
			  then selectionne k liste_inf
			  else selectionne (k-nb_inf) liste_sup
;;

let n=1000;;
let rec cree_liste = function
	|0 -> []
	|n -> (random__int (n+1))::(cree_liste (n-1))
;;

random__init (int_of_float (sys__time()));;
let test = cree_liste n;;
(* mergesort test;;
sort__sort (prefix <=) test;; *)
selectionne 10 test;;


