(** Notation postfixe **)




(* Evaluation rcursive *)
type expr = Int of int | Plus | Moins | Mult ;;

let eval_expr1 liste =
	let rec evalue = function
		|[] -> []
		|(Int a)::(Int b)::op::q -> (match op with
			|Plus -> evalue (Int (a+b)::evalue q)
			|Moins -> evalue (Int (a-b)::evalue q)
			|Mult -> evalue (Int (a*b)::evalue q)
			|_ -> evalue (Int a::evalue ((Int b)::op::q)))
		|_ as toto -> toto
in match (evalue liste) with
	|[a] -> a
	|_ -> failwith "Expression incorrecte"
;;

eval_expr1 [Int 2; Int 3;Int 4; Plus; Mult];;

(* Evaluation "acadmique", ie avec une pile *)
let eval_expr2 liste =
	let depile = function
		|Int a::Int b::q -> Int a,Int b,q
		|_ -> failwith "Erreur en dpilant"
in	let rec evalue pile = function
		|[] -> pile
		|t::q -> (match t with
			|Int _ -> evalue (t::pile) q
			|Plus -> let (Int a,Int b,m)=depile pile in evalue (Int (b+a)::m) q
			|Moins ->let (Int a,Int b,m)=depile pile in evalue (Int (b-a)::m) q
			|Mult -> let (Int a,Int b,m)=depile pile in evalue (Int (b*a)::m) q)
in match (evalue [] liste) with
	|[a] -> a
	|_ -> failwith "Expression incorrecte"
;;
(* Le filtrage n'est pas exhaustif mais c'est pas grave !!... *)

eval_expr2 [Int 2; Int 3;Int 4; Plus; Mult];;

(* Transforme une expression en arbre *)
type tree = |Feuille of int
		|Somme of tree*tree
		|Diff of tree*tree
		|Produit of tree*tree
;;

let rec arbrifie liste =
	let depile = function
		|a::b::q -> a,b,q
		|_ -> failwith "Erreur en dpilant"
in	let rec transforme pile = function
		|[] -> pile
		|t::q -> (match t with
			|Int a -> transforme (Feuille a::pile) q
			|Plus -> let (a,b,m)=depile pile in	transforme (Somme (b,a)::m) q
			|Moins ->let (a,b,m)=depile pile in transforme (Diff (b,a)::m) q
			|Mult -> let (a,b,m)=depile pile in transforme (Produit (b,a)::m) q)
in match (transforme [] liste) with
	|[a] -> a
	|_ -> failwith "Expression incorrecte"
;;

arbrifie [Int 2; Int 3;Int 4; Plus; Mult];;

(* Evaluation d'un arbre *)
let rec eval_arbre = function
	|Feuille a -> a
	|Somme   (a,b) -> (eval_arbre a)+(eval_arbre b)
	|Diff    (a,b) -> (eval_arbre a)-(eval_arbre b)
	|Produit (a,b) -> (eval_arbre a)*(eval_arbre b)
;;

eval_arbre (arbrifie [Int 2; Int 3;Int 4; Plus; Mult]);;

(* Transformation d'un arbre en expression postfixe *)
let rec arbre_to_infix = function
	|Feuille a -> string_of_int a
	|Somme   (a,b) -> (arbre_to_infix a)^" + "^(arbre_to_infix b)
	|Diff    (a,b) -> (arbre_to_infix a)^" - "^(arbre_to_infix b)
	|Produit (a,b) -> "("^(arbre_to_infix a)^") * ("^(arbre_to_infix b)^")"
;;

arbre_to_infix (arbrifie [Int 2; Int 3;Int 4; Plus; Mult]);;



