(** Plein de tris diffrents **)


(* Rend une liste de n nombres alatoires *)
random__init (int_of_float (sys__time()));;
let rec liste_alat = function 
	|0 -> []
	|n -> (random__int 10000)::(liste_alat (n-1))
;;

(* Tri par insertion d'une liste *)
let rec insert_ds_liste_trie x liste = match liste with
	|[]   -> [x]
	|t::q -> if x<=t then x::liste
			     else t::(insert_ds_liste_trie x q)
;;

let rec tri_insert = function
	|[]   -> []
	|[a]  -> [a]
	|t::q -> insert_ds_liste_trie t (tri_insert q)
;;

(* Mergesort (4 versions diffrentes) *)
(* Partie commune : la fusion des listes tries *)
let rec fusion l1 l2 = match (l1,l2) with
	|[],_		   -> l2
	|_,[]		   -> l1
	|t1::q1,t2::q2 -> if t1<t2 then t1::(fusion q1 l2) else t2::(fusion l1 q2)
;;

(* Coupe la liste en liste de listes de deux lments tries *)
let rec coupe_en_deux = function
	|[]	   -> [[]]
	|[a]	   -> [[a]]
	|a::b::q -> if a<b then [a;b]::(coupe_en_deux q) else [b;a]::(coupe_en_deux q)
;;
let mergesort1 liste = it_list fusion [] (coupe_en_deux liste);;

(* Met la liste sous forme d'une liste de listes  un lment *)
let explose liste = map (fun t->[t]) liste;;
let mergesort2 liste = it_list fusion [] (explose liste);;

(* Parcourt la liste et rend une partition de la liste en liste dj tries *)
let partition liste =
	let rec boulot en_cours rsultat = function
		|[]	-> en_cours::rsultat
		|t::q -> match en_cours with
				|[]	-> boulot [t] rsultat q
				|x::_ -> if t<=x then boulot (t::en_cours) rsultat q
						     else boulot [t] (en_cours::rsultat) q
in	boulot [] [] liste;;
let mergesort3 liste = it_list fusion [] (partition liste);;

(* Une autre vision du tri fusion : Aho-Ullmann page 85 *)
let rec egrem = function
	|[] -> ([],[])
	|[a] -> ([a],[])
	|(a::b::z) -> let (u,v)= egrem z in (a::u,b::v)
;;
let rec mergesort4 = function
	|[] -> []
	|[a] -> [a]
	|liste ->let (x,y)=egrem liste in fusion (mergesort4 x) (mergesort4 y)
;;

(* QuickSort d'une liste (plus facile qu'avec les tableaux) *)
let rec separe pivot list (inf,egal,sup) = match list with
	|[]	-> inf,egal,sup
	|t::q -> if t=pivot then separe pivot q (inf,t::egal,sup)
		   		  else if t<pivot then separe pivot q (t::inf,egal,sup)
							else separe pivot q (inf,egal,t::sup)
;;
let rec quicksort liste = match liste with
		|[]	-> []
		|t::q -> let (inf,egal,sup)=separe t liste ([],[],[])
			   in (quicksort inf)@egal@(quicksort sup)
;;

(* Tri par tas ou Heapsort *)
type 'a Tas =
	|Rien
	|Noeud of 'a * 'a Tas * 'a Tas
;;

(*let test = Noeud(7,Noeud(4,Noeud(5,Rien,Rien),Noeud(3,Rien,Rien)),Noeud(8,Noeud(9,Rien,Rien),Noeud(12,Rien,Rien)));;*)

let racine = function
	|Rien		     -> failwith "Tas vide"
	|(Noeud (x,_,_)) -> x
;;

let rec enlever = function
	|Rien			   -> Rien
	|(Noeud (x,Rien,fd)) -> fd
	|(Noeud (x,fg,Rien)) -> fg
	|(Noeud (_,fg,fd))   -> let x=racine fg and y=racine fd in
						if x<y then (Noeud (x,enlever fg,fd))
							 else (Noeud (y,fg,enlever fd))
;;

let rec ajouter y untas = match untas with
	|Rien 		 -> (Noeud (y,Rien,Rien))
	|(Noeud (x,fg,fd)) -> if x<y then (Noeud (x,fd,ajouter y fg))
					     else (Noeud (y,fd,ajouter x fg))
;;

let rec najouter liste untas = match liste with
	|[]	-> untas
	|t::q -> najouter q (ajouter t untas)
;;

let rec vider untas = match untas with
	|Rien -> []
	|tas  -> (racine tas)::(vider (enlever tas))
;;

let tripartas liste = vider (najouter liste Rien);;

(* Une petite comparaison !! *)
let liste = liste_alat 1000;;
let temp=ref (sys__time());;
tri_insert liste;;
print_float (sys__time()-. !temp); print_string " s\n"; temp:=sys__time();;
mergesort1 liste;;
print_float (sys__time()-. !temp); print_string " s\n"; temp:=sys__time();;
mergesort2 liste;;
print_float (sys__time()-. !temp); print_string " s\n"; temp:=sys__time();;
mergesort3 liste;;
print_float (sys__time()-. !temp); print_string " s\n"; temp:=sys__time();;
mergesort4 liste;;
print_float (sys__time()-. !temp); print_string " s\n"; temp:=sys__time();;
quicksort liste;;
print_float (sys__time()-. !temp); print_string " s\n"; temp:=sys__time();;
tripartas liste;;
print_float (sys__time()-. !temp); print_string " s\n"; temp:=sys__time();;
