(* Le flocon de Von Koch *)
(*************************)


(* Dfinition des oprations dans les complexes *)
type Complexe=C of (float*float);;

let Re z=match z with C(x,_)->x;;
let Im z=match z with C(_,x)->x;;
let Conj z=C(Re(z), -.Im(z));;
let Moducar z=Re(z)*.Re(z) +. Im(z)*.Im(z);;

#uninfix "plus";;
let plus x y=C(Re(x)+.Re(y), Im(x)+.Im(y));;
#infix "plus";;

#uninfix "moins";;
let moins x y=C(Re(x)-.Re(y), Im(x)-.Im(y));;
#infix "moins";;

#uninfix "fois";;
let fois x y=C(Re(x)*.Re(y)-.Im(x)*.Im(y), Re(x)*.Im(y)+.Re(y)*.Im(x));;
#infix "fois";;

#uninfix "divi";;
let divi x y= let a=x fois (Conj y) in C(Re(a)/.(Moducar y),Im(a)/.(Moducar y));;
#infix "divi";;

(* ici Arg(z) est dans ]-Pi;Pi] *)
let Arg x = let z=x divi (C(sqrt(Moducar x),0.)) in
	if Re(z)= -1. then 4.*.atan(1.)
	else 2.*.atan(Im(z)/.(Re(z)+.1.))
;;

(* Dfinitions des transformations *)
#open "graphics";;

let valeurs a = [|C(0.,0.);C(a,0.);C(0.5,sqrt(a-.0.25));C(1.-.a,0.);C(1.,0.)|];;

let S alpha = let a=valeurs alpha and result = make_vect 4 Conj
	in	for j=0 to 3 do
			result.(j)<-(function z -> (a.(j+1) moins a.(j))fois z plus a.(j))
		done;
		result;;

(* Von Koch -> 1/3; Segment -> 1/4; Triangle -> 1/2; Essayez 0.4 ! *)
let similitudes=S (1./.3.);;
let S_0 = similitudes.(0);;
let S_1 = similitudes.(1);;
let S_2 = similitudes.(2);;
let S_3 = similitudes.(3);;

(* Trac avec des itrations de similitudes *)
let rec calcul = function
	| [] -> []
	| t::q -> (S_0 t)::(S_1 t)::(S_2 t)::(S_3 t)::(calcul q)
;;

let convert t = int_of_float (1000.*.t)+10;; (*  adapter  la rsolution de l'affichage *)
let affiche (C(x,y)) = plot (convert x) (convert y);;
let trace = do_list affiche;;
	
let rec iteration liste = function
	| 0 -> trace liste
	| n -> ((*trace liste;*) iteration (calcul liste) (n-1))
;;

open_graph ""; iteration [C(0.,0.);C(1.,0.)] 6;;

(* Trac avec des convergences de suites de fonctions *)
let ligne (C(x,y)) = lineto (convert x) (convert y);;

let trace_fct f =
	moveto 10 10;
	for t=0 to 4999 do
		ligne (f (float_of_int t /.5000.));
	done
;;

let U f = function t -> 
	let j=int_of_float (4.*.t) in
		if j=4
		then similitudes.(3)(f(4.*.t-.float_of_int j))
		else similitudes.(j)(f(4.*.t-.float_of_int j));;

let rec iteration_fct f = function
	| 0 -> trace_fct f 
	| n -> ((*trace_fct f;*) iteration_fct (U f) (n-1))
;;

open_graph ""; iteration_fct (function t -> C(t,0.)) 6;;

(* Dfinissons de la mme faon le triangle de Serpinski *)
let similitudes=[|(function z -> C(0.5,0.) fois z);
						(function z -> C(0.5,0.) fois z plus C(0.5,0.));
						(function z -> C(0.5,0.) fois z plus C(0.25,(sqrt 3.)/.4.))|];;
let S_0 = similitudes.(0);;
let S_1 = similitudes.(1);;
let S_2 = similitudes.(2);;

(* Trac avec des itrations de similitudes *)
let rec calcul = function
	| [] -> []
	| t::q -> (S_0 t)::(S_1 t)::(S_2 t)::(calcul q)
;;

let convert t = int_of_float (800.*.t)+10;; (*  adapter  la rsolution de l'affichage *)
let affiche (C(x,y)) = plot (convert x) (convert y);;
let trace = do_list affiche;;

let rec iteration liste = function
	| 0 -> trace liste
	| n -> ((*trace liste;*) iteration (calcul liste) (n-1))
;;

open_graph ""; iteration [C(0.,0.);C(1.,0.);C(0.5,(sqrt 3.)/.2.)] 8;;
